/* ----------------------------------------------- */
/* FormRead  -  WWW form reader                    */
/*                                                 */
/* Copyright 1997-2002 Simone Zanella Productions   */
/* Permission is granted to redistribute freely    */
/* and use for any purpose, commercial or private, */
/* as long as this copyright notice is retained    */
/* and the source code is included free of charge  */
/* with any binary.                                */
/*                                                 */
/* Description:                                    */
/*                                                 */
/* This program allow to read a form received as:  */
/*                                                 */
/*       application/x-www-form-urlencoded         */
/*                                                 */
/* by writing a temporary file based on a template */
/* file which must reside in the same directory as */
/* FormRead and has name Form_Name (hidden field   */
/* in form data) with extension ".frt".            */
/* This file is passed as parameter to the appli-  */
/* cation whose pathname is stored into            */
/* Form_Name+".frr" (field READER); this config    */
/* file can also include additional command line   */
/* parameters for FormRead (field FORCEPAR).       */
/*                                                 */
/* If file ".frt" does not exist, a text (or HTML, */
/* if parameter -h is specified) file is created,  */
/* with this layout:                               */
/*             field name  value                   */
/*                                                 */
/* Each line is terminate by CR+LF (-t, default),  */
/* CR (parameter -m) or LF (parameter -u).         */
/* Template file cannot contain null characters    */
/* (ascii 0).                                      */
/*                                                 */
/* If field Form_Name does not exist, the reader   */
/* stored in 'default.frr' is used on the standard */
/* text layout.                                    */
/*                                                 */
/* If you configure FormRead as helper application */
/* in Netscape->General Preferences->Helpers, by   */
/* adding the type:                                */
/*                                                 */
/*      Mime type: application                     */
/*      Mime subtype: x-www-form-urlencoded        */
/*                                                 */
/* and setting "Launch the application" to the     */
/* pathname of FormRead, by selecting a mimed form */
/* received you will be able to read the form      */
/* in human readable format, with the viewer you   */
/* chose (which can be the same Netscape, if you   */
/* use parameter -h or the template is in HTML).   */
/*                                                 */
/* Errors can only happen when data file is        */
/* opened; in this case, an error message is       */
/* printed and displayed with the viewer,          */
/* reporting the kind of error. All other errors   */
/* are simply ignored (but should never happen,    */
/* on normal conditions).                          */
/*                                                 */
/* If KEEP_DATA is #defined, temp file is not re-  */
/* moved, but it is overwritten at each call, is   */
/* located in FormRead directory and is called     */
/* 'formdata.txt' (or 'formread.htm', if flag -h   */
/* is specified, or 'formdata'+extension specified */
/* with parameter -e).                             */
/* This is useful under Windows 95, because when   */
/* you run a Windows app from a Dos app the Dos    */
/* app does NOT wait for the Windows app to end    */
/* before continuing.                              */
/*                                                 */
/* If UNIX_PATH is #defined, then slash character  */
/* is used as path separator instead of backslash. */
/*                                                 */
/* Written by Simone Zanella.                      */
/*            <szp@zanella-hifi.com>               */
/*                                                 */
/* Home Page:                                      */
/* <http://www.zanella-hifi.com/szp>               */
/*                                                 */
/* Conversion routines:                            */
/*            htoi, url_unescape and ishex         */
/*                                                 */
/* were taken from UNCGI v1.8 by                   */
/* Steven Grimm <koreth@hyperion.com>              */
/*                                                 */
/* UNCGI is available from:                        */
/* <http://www.hyperion.com/~koreth/uncgi.html>    */
/*                                                 */
/* Last modified:   23/10/1997                     */
/* 23/10   now file extension for temporary file   */
/*         is '.htm' if flag -h is specified       */
/* 4/12    changed '.frr' file format; added       */
/*         parameters '-e', '-t', '-n'             */
/*                                                 */
/* ----------------------------------------------- */
/* SYNTAX:                                         */
/* FormRead [-u | -m | -t] [-h | -n] [-o dest]     */
/*          [-d dir] [-e ext] file                 */
/*                                                 */
/* -o   Save report to dest: reader is not invoked */
/* -d   Save report in dir: a unique file name is  */
/*      used with the mask 'formXX.XXX'            */
/* -e   Set output file extension; default is      */
/*      '.txt' for text files and '.htm' if HTML   */
/*      filter is active                           */
/*                                                 */
/* -u   Unix format (lines end with CR)            */
/* -m   Macintosh format (lines end with LF)       */
/* -t   Dos format (lines end with CR+LF)          */
/* -h   HTML filter                                */
/* -n   disable HTML filter                        */
/*                                                 */
/* Parameters -u/-m/-t and -h/-n make sense only   */
/* for those forms which have no template file and */
/* determine the format of the report created,     */
/* which is by default a standard Dos text file    */
/* (without HTML filter).                          */
/*                                                 */
/* ----------------------------------------------- */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* -------------------------------- Headings ------------------------------- */

#define INTESTA "FormRead - FormReader v%s / (C) %s Simone Zanella Productions\n"
#define INTESTB "PC Software Development - http://www.zanella-hifi.com/szp\n"

/*
** If KEEP_DATA is #defined, use file 'formdata.txt'/'formdata.htm' as
** temporary file: it is overwritten at each call, but is not removed
** when program ends.
*/

/*
** If UNIX_PATH is #defined, use '/' insted of '\' as path separator token.
** PATH_SEP is set automatically to the correct type with compilers
** DJGPP and Borland C. KEEP_DATA is automatically defined in the same
** circumstances.
*/

#ifdef __DJGPP__
  #define PATH_SEP  '\\'
  #define KEEP_DATA
#else
  #ifdef __BORLANDC__
    #include <dir.h>
    #define PATH_SEP  '\\'
    #define KEEP_DATA
  #else
    #ifdef UNIX_PATH
      #define PATH_SEP  '/'
    #else
      #define PATH_SEP  '\\'
      #define KEEP_DATA
    #endif
  #endif
#endif

/* -------------------------- Version information -------------------------- */

#define VERS      "1.2"
#define SIGNATURE "FormRead 1.2"
#define YEAR      "2002"

/* ---------------------------- Pseudo-functions --------------------------- */

#define ishex(x) (((x) >= '0' && (x) <= '9') || ((x) >= 'a' && (x) <= 'f') || \
                  ((x) >= 'A' && (x) <= 'F'))

/* -------------------------- Program definitions -------------------------- */

#define MAXFIELD   512           /* Maximum number of fields in data file */
#define FORM_NAME  "Form_Name"   /* Field for template and reader file */
#define EXT_CFG    ".frr"        /* Configuration file extension */
#define EXT_TEMP   ".frt"        /* Template file extension */

/* --------------------------- Public variables ---------------------------- */

static int  out_type = 0;     /* Standard text, for forms without template */
static char pathname[128];    /* Scratch space for path operations */
static char basepath[128];    /* Base path for template and reader files */
static int  lp;               /* Length of base path */
static char eol[3];           /* End of line characters */
static int remfile = 1;       /* Start reader and remove file on exit */
static char *tempfile = NULL; /* Temporary file name */
static char *filename = NULL; /* Name of file to process */
static char *outext = NULL;   /* Extension for output file */

typedef struct fieldval {
  char *field;
  char *value; } FieldVal;

static FieldVal fieldtable[MAXFIELD];  /* Field-value table */

static int numfield = 0;   /* Number of fields in fieldtable */
static int fnpos = -1;     /* Position of FORM_NAME in fieldtable (-1 = absent) */


/* -------------------------- Auxiliary functions -------------------------- */

static char *skipblank(char * string)
{
  /*
  ** Return a pointer to the next character different from a space.
  */

  while (*string == ' ')
    string++;

  return string;
}

static char *rtrim(char * string)
{
  /*
  ** Remove trailing spaces from string.
  */

  register char * pp;

  if (*string) /* Not a null string */
  {
    pp = string + strlen(string) - 1; /* Last character */
    while (pp > string && *pp == ' ') /* Find first not-space */
      pp--;
    if (pp == string) /* Only one character: is space? */
      if (*string == ' ')
        *string = 0;
      else
        *++pp = 0;
    else
      *++pp = 0;
  }
  return string;
}

static char *ltrim(char * string)
{
  /*
  ** Remove spaces at the beginning of string by moving the string
  ** to the left. Memory is not reallocated.
  */

  register char * cp;
  register long l;

  cp = skipblank(string);
  if (*cp)
  {
    l = strlen(cp);
    memmove(string, cp, l + 1);
  }
  else /* Empty string */
    *string = 0;

  return string;
}

static void *checknul(void *punt)
{
  /*
  ** Stop program if allocation function failed.
  */

  if (punt == NULL)
    exit(0);
  return punt;
}

#define BUFSIZE    4096
#define MAXIDLEN     80

static int listread(const char *listfile, const char *id, char **pstring)
{
  /*
  ** Return in *pstring (allocated) the contents of the line corresponding
  ** to id in listfile.
  **
  ** Returns 0 if:
  ** - listfile does not exist;
  ** - id does not exist within listfile;
  ** - listfile cannot be opened.
  **
  ** Returns -1 if:
  ** - listfile contains a line too long (>BUFSIZE).
  **
  ** Returns 1 if id was found in listfile and copies the corresponding line
  ** into *pstring (newly allocated).
  **
  ** Format for listfile:
  **
  ** id1: text
  ** id2: text
  ** ..
  */

  char *inbuf, *idcopy;
  FILE *fdin;
  int retval = 0, firsttime = 1;
  long offset, count, l;
  register char *sp, *cp;
  char *newstring = NULL;

  idcopy = checknul(malloc(MAXIDLEN + 3));

  strcpy(idcopy, id);
  strcat(idcopy, ":");

  /* Open listfile */
  inbuf = checknul(malloc(BUFSIZE));

  fdin = fopen(listfile, "rt");

  if(fdin == NULL)
  {
    free(inbuf);
    free(idcopy);
    return 0;
  }

  /* Determine text format */

  offset = 0;
  while (1)
  {
    /* Load into inbuf part of listfile */
    count = fread(inbuf + offset, 1, BUFSIZE - 1 - offset, fdin);

    /* Insert chr(0) at the end, to allow string operations */
    *(inbuf + offset + count) = 0;

     /* Find idcopy in inbuf */
    sp = strstr(inbuf, idcopy);

    /* Not found: remove the part of the buffer before next line */
    if (sp == NULL || (firsttime && sp != inbuf && *(sp - 1) != '\n'))
    {
      if (firsttime)
      {
        firsttime = 0;
        strcpy(idcopy, "\n");
        strcat(idcopy, id);
        strcat(idcopy, ":");
      }

      if (count == (BUFSIZE - 1 - offset)) /* Not EOF */
      {
        /* Keep only last uncomplete line */
        sp = strrchr(inbuf, '\n');
        if (sp)
        {
          offset = inbuf + BUFSIZE - 1 - sp;
          memmove(inbuf, sp, offset);
        }
        else /* String too long */
        {
          retval = -1;
          goto error;
        }
      }
      else /* EOF: idcopy not found */
        goto error;
    }
    else
    {
      /* Found idcopy: copy line */
      sp += strlen(idcopy);

      /* Line is unterminated */
      if ((cp = strchr(sp, '\n')) == NULL)
      {
        /* Copy until the end of the buffer */
        l = strlen(sp);
        newstring = checknul(malloc(l + 1));
        strcpy(newstring, sp);
        if (count == (BUFSIZE - 1 - offset)) /* Not EOF */
        {
          /*
          ** Read and copy into the string associated to idcopy
          ** the contents of the file, until the next line.
          */
          while (1)
          {
            count = fread(inbuf, 1, BUFSIZE - 1, fdin);
            *(inbuf + count) = 0;
            sp = strchr(inbuf, '\n');
            if (sp != NULL) /* EOL found */
            {
copy:
              /* Copy until the end of line */
              *sp = 0;
              newstring = checknul(realloc(newstring, l + (sp - inbuf) + 1));
              strcpy(newstring + l, inbuf);
              goto out;
            }
            else
              if (count < BUFSIZE - 1) /* EOF */
              {
                sp = inbuf + count;
                goto copy;
              }
              else /* Not EOF, but EOL not read yet */
              {
                /* Copy all the buffer and continue */
                newstring = checknul(realloc(newstring, l + count + 1));
                strcpy(newstring + l, inbuf);
                l += count;
              }
          }
        }
        else /* End of file: newstring contains the string wanted */
          goto out;
      }
      else /* Terminated line: copy and exit */
      {
        *cp = 0;
        l = cp - sp;
        if (l == 0)
          newstring = checknul(strdup(""));
        else
        {
          newstring = checknul(malloc(l + 1));
          strcpy(newstring, sp);
        }
        goto out;
      }
    }
  }
out:
  fclose(fdin);
  *pstring = ltrim(rtrim(newstring));
  free(inbuf);
  free(idcopy);
  return 1;

error:
  fclose(fdin);
  free(newstring);
  free(inbuf);
  free(idcopy);
  return retval;
}

static long flength(FILE *fp)
{
  /* Return file length (all files are opened in binary mode) */

  fpos_t pos;
  register long size;

  fgetpos(fp, &pos);
  fseek(fp, 0, 2);
  size = ftell(fp);
  fsetpos(fp, &pos);
  return size;
}

static void wrline(FILE *fp, char *buffer)
{
  /* Write a line of text and append end-of-line */

  fwrite(buffer, 1, strlen(buffer), fp);
  fwrite(eol, 1, strlen(eol), fp);
}

static void wrtext(FILE *fp, char *buffer)
{
  /* Write text to file */

  fwrite(buffer, 1, strlen(buffer), fp);
}

static void inittemp(FILE *fp, char *title)
{
  /* Prepare standard layout (no template file), both text and html */

  if (out_type) /* HTML */
  {
    wrtext(fp, "<html><head><title>");
    wrtext(fp, title);
    wrtext(fp, "</title></head><body><h1>");
    wrtext(fp, title);
    wrline(fp, "</h1><table border=\"3\" cellpadding=\"3\" width=\"100%\">");
  }
  else
  {
    wrline(fp, title);
    wrline(fp, "");
  }
}

static void closetemp(FILE *fp)
{
  /* Close standard layout (no template file), both text and html */

  if (out_type) /* HTML */
  {
    wrtext(fp, "</table><br><br><strong>Converted by <cite>");
    wrtext(fp, SIGNATURE);
    wrline(fp, "</cite></strong></body></html>");
  }
  else
  {
    wrline(fp, "");
    wrtext(fp, "Converted by ");
    wrline(fp, SIGNATURE);
  }
}

static void error(FILE *fp, char *message)
{
  /* Print error for standard layout (no template file), both text and html */

  inittemp(fp, "Error!");
  if (out_type) /* HTML */
    wrtext(fp, "<tr><td><h2>");
  wrtext(fp, message);
  if (out_type) /* HTML */
    wrline(fp, "</h2></td></tr>");
  closetemp(fp);
  fclose(fp);
}

static void stdzpath(char *cfgfile, char *extension)
{
  /*
  ** Copy cfgfile at the end of pathname, which keeps the full path
  ** of the executable.
  */

  char i;

  i = lp;
  strcpy(pathname, basepath);
  while (*cfgfile) pathname[i++] = *cfgfile++;
  while (*extension) pathname[i++] = *extension++;
  pathname[i] = 0;
}

static int htoi(unsigned char *s)
{
  /*
  ** Convert two hex digits to a value.
  ** This routine is taken from UNCGI v1.8.
  */

  int  value;
  char c;

  c = s[0];
  if (isupper(c))
    c = tolower(c);
  value = (c >= '0' && c <= '9' ? c - '0' : c - 'a' + 10) * 16;

  c = s[1];
  if (isupper(c))
    c = tolower(c);
  value += c >= '0' && c <= '9' ? c - '0' : c - 'a' + 10;

  return (value);
}

static void url_unescape(unsigned char *str)
{
  /*
  ** Get rid of all the URL escaping in a string.  Modify it in place, since
  ** the result will always be equal in length or smaller.
  ** This routine is taken from UNCGI v1.8.
  */

  unsigned char *dest = str;

  while (str[0])
  {
    if (str[0] == '+')
      dest[0] = ' ';
    else if (str[0] == '%' && ishex(str[1]) && ishex(str[2]))
    {
      dest[0] = (unsigned char) htoi(str + 1);
      str += 2;
    }
    else
      dest[0] = str[0];

    str++;
    dest++;
  }

  dest[0] = '\0';
}

static void scanquery(char *q)
{
  /*
  ** Scan a query string, building field/value table.
  */

  char *next = q, *equal;

  do
  {
     next = strchr(next, '&');
     if (next)
       *next = '\0';

     url_unescape(q);

     /* Add to field table */

     equal = strchr(q, '=');

     if (equal)
       *equal++ = 0;

     /* Store field name, length and value */

     fieldtable[numfield].field = rtrim(q);
     if (strcmp(q, FORM_NAME) == 0)
       fnpos = numfield;
     fieldtable[numfield++].value = rtrim(equal);

     if (next)
       next++;

     q = next;

  } while (q != NULL);
}

static char * strtran(char * string, const char * find, const char * replace)
{
  /*
  ** Reallocate string, so that all occurrences of 'find' can be replaced
  ** by 'replace'; substitute all occurrences and return new string.
  */

  char *newstring, *pp, *cp;
  int  s, f, r, offset;

  /* Length for various strings */

  s = strlen(string);
  f = strlen(find);
  r = strlen(replace);

  /* Find first occurrence */

  pp = strstr(string, find);

  if (!pp) /* No occurrence: return */
    return string;

  /* Length of transformed string */

  while (pp)
  {
    s += r - f;
    pp = strstr(pp + 1, find);
  }

  /* Allocate new string */

  newstring = checknul(malloc(s + 1));

  /* Copy transformed string */

  pp = strstr(string, find);
  cp = string;
  offset = 0;
  while (pp)
  {
    /* Copy everything before 'find' */

    s = pp - cp;
    memcpy(newstring + offset, cp, s);
    offset += s;

    /* Copy 'replace' */

    memcpy(newstring + offset, replace, r);
    offset += r;

    /* Process rest of string */

    pp += f;
    cp = pp;
    pp = strstr(cp, find);
  }

  /* Copy what is left */

  s = strlen(cp);
  memcpy(newstring + offset, cp, s);
  offset += s;

  /* Terminate the string */

  *(newstring + offset) = 0;

  /* Free previous string */

  free(string);

  /* Return new string */

  return newstring;
}

static void header(unsigned int npar)
{
  /* Print out informations about program, when invoked without parameters. */

  if (npar == 1)
  {
    fprintf(stderr, INTESTA, VERS, YEAR);
    fprintf(stderr, INTESTB);
    fprintf(stderr, "Use:  FormRead [-u | -m | -t] [-h | -n] [-o dest | -d dir] [-e ext] filename\n");
    fprintf(stderr, "Desc:\n");
    fprintf(stderr, "      Read a form in 'application/x-www-form-urlencoded' format and\n");
    fprintf(stderr, "      create a temporary file (optionally from a template), that is\n");
    fprintf(stderr, "      viewed through an external application.\n");
    fprintf(stderr, "      Field 'Form_Name' from data file is used to determine which\n");
    fprintf(stderr, "      template must be used (file in the same directory as FormRead\n");
    fprintf(stderr, "      with '%s' extension) and which reader to use (file with\n", EXT_TEMP);
    fprintf(stderr, "      '%s' extension). If no template is specified, a standard\n", EXT_CFG);
    fprintf(stderr, "      layout name=value is used and the file is created as a Dos\n");
    fprintf(stderr, "      text file (-t), Unix text file (parameter -u) or Macintosh text\n");
    fprintf(stderr, "      file (parameter -m); the file is in HTML if parameter -h is\n");
    fprintf(stderr, "      specified (default: -n, no HTML filtering). If no reader file is found,\n");
    fprintf(stderr, "      the reader in default%s is used. To read form data received\n", EXT_CFG);
    fprintf(stderr, "      by e-mail you must configure FormRead as helper in your browser\n");
    fprintf(stderr, "      for type 'application' and subtype 'x-www-form-urlencoded' and\n");
    fprintf(stderr, "      create a template file and viewer file for each form that you\n");
    fprintf(stderr, "      want specially formatted. To permanently save formatted data to\n");
    fprintf(stderr, "      a certain file, use '-o' or '-d' (reader is not invoked).\n");
    fprintf(stderr, "      If -e is not specified, output extension is '.txt' or '.htm'.\n");
    exit(0);
  }
  return;
}

static int sortlength(const void * f1, const void * f2)
{
  /* Sort field names by decreasing length */

  return ( strlen(((FieldVal *) f2)->field) - strlen(((FieldVal *) f1)->field) );
}

static int processopt(const char *current, const char *next)
{
  if (*current == '-')
  {
    switch (*(current + 1))
    {
      case 't': case 'T': /* Dos output */
        strcpy(eol, "\x0D\x0A");
        break;
      case 'u': case 'U': /* Unix output */
        strcpy(eol, "\x0A");
        break;
      case 'm': case 'M': /* Mac output */
        strcpy(eol, "\x0D");
        break;
      case 'n': case 'N': /* Disable Html filter */
        out_type = 0;
        break;
      case 'h': case 'H': /* Html filter */
        out_type = 1;
        break;
      case 'o': case 'O': /* Destination is not temporary */
        free(tempfile);
        tempfile = checknul(strdup(next));
        remfile = 0;
        return 2;
      case 'e': case 'E': /* Output extension */
        free(outext);
        outext = checknul(strdup(next));
        return 2;
      case 'd': case 'D': /* Save form to specified directory */
        {
          int j;

          free(tempfile);
          tempfile = malloc(strlen(next) + 2 + 6);
          strcpy(tempfile, next);
          j = strlen(tempfile);

          /* Add path separator, if missing */
          if (*(tempfile + j - 1) != PATH_SEP)
          {
            *(tempfile + j) = PATH_SEP;
            *(tempfile + j + 1) = 0;
          }

          /* Add template string */
          strcat(tempfile, "formXXXXXX");
          mktemp(tempfile);
          remfile = 0;
        }
        return 2;
      default:
        fprintf(stderr, "FormRead: unknown parameter \"%s\" ignored.", current);
        break;
    }
    return 1;
  }
  else
  {
    free(filename);
    filename = checknul(strdup(current));
    return 0;
  }
}

void main(int argc, char *argv[])
{
  FILE *fp, *df, *tp; /* Temp file, data file, template file */
  char *cp, *cpe, *basename, *reader = NULL, *forcepar = NULL;
  unsigned int i;

  header(argc);
  filename = NULL;

  /* Save base path */
  cp = cpe = argv[0];
  while (*cpe != 0) cpe++;
  while (*cpe != PATH_SEP && cpe > argv[0]) cpe--;
  lp = 0;
  if (*cpe == PATH_SEP)
  {
    cpe++;
    while (cp < cpe) pathname[lp++] = *cp++;
  }
  pathname[lp] = 0;
  strcpy(basepath, pathname);

  /* Process command line parameters */
  strcpy(eol, "\x0D\x0A");
  for (i = 1; i < argc;)
    switch (processopt(argv[i], (i + 1 < argc) ? argv[i + 1] : NULL))
    {
      case 0:
        goto endclpar;
      case 1:
        i++;
        break;
      case 2:
        i += 2;
        break;
    }
endclpar:

  /* Check existance of data file */
  if (filename == NULL)
  {
    error(fp, "Data file not specified.");
    goto noreader;
  }

  /* Open data file */
  df = fopen(filename, "rb");
  if (df == NULL)
  {
    error(fp, "Could not open data file.");
    goto noreader;
  }

  /* Copy data file to memory */
  i = flength(df);
  cp = malloc(i + 1);
  fread(cp, 1, i, df);
  *(cp + i) = 0;

  /* Build field table */
  scanquery(cp);

  /* If special processing is requested.. */
  basename = (fnpos == -1) ? NULL : checknul(strdup(fieldtable[fnpos].value));
  if (basename)
  {
    /* Custom reader (if found); revert to standard reader on error or not found */
    stdzpath(basename, EXT_CFG);

    listread(pathname, "FORCEPAR", &forcepar);
    if (listread(pathname, "READER", &reader) == 0)
      goto noreader;
  }
  else
  {
    /* Standard reader: default+EXT_CFG MUST always be present in FormRead directory! */
noreader:

    stdzpath("default", EXT_CFG);

    listread(pathname, "FORCEPAR", &forcepar);
    if (listread(pathname, "READER", &reader) == 0)
      goto end;
  }

  /* Process additional parameters */
  if (forcepar)
  {
    cp = strtok(forcepar, " ");
    cpe = strtok(NULL, " ");
    while (cp)
    {
      processopt(cp, cpe);
      cp = strtok(NULL, " ");
      cpe = strtok(NULL, " ");
    }
  }

  /* Prepare temp file name, fully qualified */
  if (tempfile == NULL)
  {
  #ifdef KEEP_DATA
    stdzpath("formdata", (outext) ? outext : ((out_type) ? ".htm" : ".txt"));
    tempfile = checknul(strdup(pathname));
  #else
    tempfile = tempnam(pathname, "fr");
  #endif
  }

  /* Open temp file */
  fp = fopen(tempfile, "wb");

  if (basename)
  {
    /* Open template (if found); revert to standard template on error */
    stdzpath(basename, EXT_TEMP);

    tp = fopen(pathname, "rb");
    if (tp == NULL)
      goto notemplate;

    /* Copy template file to memory */
    i = flength(tp);
    cp = malloc(i + 1);
    fread(cp, 1, i, tp);
    fclose(tp);
    *(cp + i) = 0;

    /* Sort by decreasing length field names, to avoid problems like:
    **
    ** My_Field
    ** My_Field2  <-  replaced by value associated to My_Field, followed by 2,
    **                if My_Field comes before My_Field2 in data file
    */

    qsort(fieldtable, numfield, sizeof(fieldtable[0]), sortlength);

    /* Translate template */
    for (i = 0; i < numfield; i++)
      cp = strtran(cp, fieldtable[i].field, fieldtable[i].value);

    /* Write transformed template to temp file */
    fwrite(cp, 1, strlen(cp), fp);
  }
  else
  {
notemplate:

    /* Write data in standard layout (name=value) */
    inittemp(fp, "Form results");
    for (i = 0; i < numfield; i++)
    {
      if (out_type) /* HTML */
        wrtext(fp, "<tr><td><strong>");
      wrtext(fp, fieldtable[i].field); /* Field name */
      if (out_type) /* HTML */
        wrtext(fp, "</strong></td><td><cite>");
      else
        wrtext(fp, ": ");
      wrtext(fp, fieldtable[i].value); /* Field value */
      if (out_type) /* HTML */
        wrline(fp, "</cite></td></tr>");
      else
        wrline(fp, "");
    }
    closetemp(fp);
  }

  /* Close temp file */
  fclose(fp);

  /* Reader invoked only if parameter -o not specified */
  if (remfile)
  {
    /* Create string to run reader */
    cpe = malloc(strlen(tempfile) + strlen(reader) + 2);
    strcpy(cpe, reader);

    /* Concatenate temp file name */
    strcat(cpe, " ");
    strcat(cpe, tempfile);

    /* Call reader on temp file */
    system(cpe);

end:
#ifndef KEEP_DATA
    /* Remove temporary file */
    remove(tempfile);
#endif
  }
}
